import numpy as np
import scipy.io as scio
import cv2
'''
this file have some code about pre-processing some image
'''

# 读取mat文件
def readMatFile(matFilePath):
    matFile = scio.loadmat(matFilePath)
    return matFile
    
# 使用opencv读取文件并且转化成RGB图像
def cv2Bgr2Rgb(imgPath):
    img = cv2.imread(imgPath)
    img = cv2.cvtColor(img,cv2.COLOR_BGR2RGB)
    return img


    
# -------------------------------------------------------下面是brp-net里面用到的一些函数---------------------------------------------------------
# 图像的变形        
def transformImg(img):
    norm_mean=[0.485, 0.456, 0.406]
    norm_std=[0.229, 0.224, 0.225]

    _, _, mod = img.shape
    img = Image.fromarray(img.astype(np.uint8))
    img = ImageEnhance.Color(img).enhance(seed())
    img = ImageEnhance.Brightness(img).enhance(seed())
    img = ImageEnhance.Contrast(img).enhance(seed())
    img = ImageEnhance.Sharpness(img).enhance(seed())
    img = np.asarray(img).astype(np.float32)
    img = img.transpose([2, 0, 1])
    for imod in list(range(mod)):
        img[imod] = (img[imod]/255.0 - norm_mean[imod])/norm_std[imod]
    img += np.random.normal(0, np.random.rand(), img.shape)*0.01
    return img

def test_extract_patches(img, patch_sz, stride):
    b, c,  h, w = img.shape
    padding_h = (patch_sz - stride)//2
    padding_w = (patch_sz - stride)//2
    padding_h_img = int(np.ceil(h/stride)*stride - h) // 2
    padding_w_img = int(np.ceil(w/stride)*stride - w) // 2
    pad_img = F.pad(img, (padding_w_img + padding_w, padding_w_img + padding_w, padding_h_img + padding_h, padding_h_img + padding_h), mode='reflect')
    return pad_img

# 一些提取图片的操作
# 里面包含了一些重叠采样的方法
# 0 - 256 - 0 256
# 0 - 256 - 128 384
# 0 - 256 - 256 512
# 0 - 256 - 384 640
# 0 - 256 - 512 768
# 0 - 256 - 640 896
# 0 - 256 - 768 1024
# 0 - 256 - 896 1152
def extract_patches(img, patch_sz, stride):
    b, c,  h, w = img.shape
    padding_h = (patch_sz - stride)//2
    padding_w = (patch_sz - stride)//2
    padding_h_img = int(np.ceil(h/stride)*stride - h) // 2
    padding_w_img = int(np.ceil(w/stride)*stride - w) // 2
    # reflect将右边的值镜像过来
    pad_img = F.pad(img, (padding_w_img + padding_w, padding_w_img + padding_w, padding_h_img + padding_h, padding_h_img + padding_h), mode='reflect')
    # [4, 3, 1152, 1152]
    _, _, h_pad, w_pad = pad_img.shape
    patches = []
    ibs = []
    shs = []
    sws = []
    imgCount = 0
    for ib in list(range(b)):
        for sh in list(range(padding_h, padding_h+h+padding_h_img*2, stride)):
            for sw in list(range(padding_w, padding_w+w+padding_w_img*2, stride)):
                tmp_p = pad_img[ib, :, (sh-padding_h):(sh+padding_h+stride), (sw-padding_w):(sw+padding_w+stride)].unsqueeze(dim=0)
                patches.append(tmp_p)
                ibs.append(ib)
                shs.append(sh)
                sws.append(sw)
    patches = torch.cat(tuple(patches), dim=0)
    return patches, ibs, shs, sws

# 从patch中重构图片
def reconstruct_from_patches_weightedall(patches, ibs, shs, sws, patch_sz, stride, b, c, h, w, patches_weight_map):
    padding_h = (patch_sz - stride)//2
    padding_w = (patch_sz - stride)//2
    padding_h_img = int(np.ceil(h/stride)*stride - h) // 2
    padding_w_img = int(np.ceil(w/stride)*stride - w) // 2
    img_rc = torch.zeros(b, c, h+2*padding_h_img+2*padding_h, w+2*padding_w_img+2*padding_w)
    ncount = torch.zeros(b, c, h+2*padding_h_img+2*padding_h, w+2*padding_w_img+2*padding_w)
    #if len(patches_weight_map.shape)==3:
    #    patches_weight_map = patches_weight_map.unsqueeze(dim=0)
    ipatches = 0
    for ipatches in list(range(patches.shape[0])):
        ib = ibs[ipatches]
        sh = shs[ipatches]
        sw = sws[ipatches]
        img_rc[ib, :, (sh-padding_h):(sh+padding_h+stride), (sw-padding_w):(sw+padding_w+stride)] += patches[ipatches] * patches_weight_map
        ncount[ib, :, (sh-padding_h):(sh+padding_h+stride), (sw-padding_w):(sw+padding_w+stride)] += patches_weight_map
    img_rc_norm =  img_rc / ncount
    img_rc_norm = img_rc_norm[:, :, (padding_h_img+padding_h):(padding_h_img+padding_h+h), (padding_w_img+padding_w):(padding_w_img+padding_w+w)]
    return img_rc_norm








import time 
from PIL import Image, ImageEnhance
import numpy as np
import torch.nn.functional as F 
import torch
import os
import time
import scipy.io as scio
import random
import itertools
from scipy import misc
import torch.nn as nn
from torch.autograd import Variable
import torch.utils.data as data
import math
from scipy.ndimage.filters import gaussian_filter, median_filter
import multiprocessing



# 获取当前格式化的时间
def getNowTime(formatter = '%Y-%m-%d_%H_%M_%S'):
    return time.strftime(formatter, time.localtime())

def seed():
    return np.random.rand()*0.1+0.9







    










import numpy as np
from skimage.morphology import label
from scipy import ndimage
import scipy.io as scio
import cv2

def get_rect_of_mask(img):
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]
    return rmin, rmax, cmin, cmax
    
def get_size_of_mask(img):
    if np.max(img) == 0:
        return 0
    rmin, rmax, cmin, cmax = get_rect_of_mask(img)
    return max([rmax - rmin, cmax - cmin])
    
def remove_overlaps(instances, scores):
    if len(instances) == 0:
        return [], []
    lab_img = np.zeros(instances[0].shape, dtype=np.int32)
    for i, instance in enumerate(instances):
        lab_img = np.maximum(lab_img, instance * (i + 1))
    instances = []
    new_scores = []
    for i in range(1, lab_img.max() + 1):
        instance = (lab_img == i).astype(np.bool)
        if np.max(instance) == 0:
            continue
        instances.append(instance)
        new_scores.append(scores[i - 1])
    return instances, new_scores
    
def post_proc(output, cutoff=0.5, cutoff_instance_max=0.3, cutoff_instance_avg=0.2, post_dilation_iter=2, post_fill_holes=True):
    """
    Split 1-channel merged output for instance segmentation
    :param cutoff:
    :param output: (h, w, 1) segmentation image
    :return: list of (h, w, 1). instance-aware segmentations.
    """
    # The post processing function 'post_proc' is borrowed from the author of CIA-Net.
    
    cutoffed = output > cutoff
    # 标记整数数组的连接区域,获取连通区域
    lab_img = label(cutoffed, connectivity=1)
    instances = []
    # pdb.set_trace()
    for i in range(1, lab_img.max() + 1):
        #将每个连通分类对应的单独的图片提取出来
        instances.append((lab_img == i).astype(np.bool))

    filtered_instances = []
    scores = []
    for instance in instances:
        # TODO : max or avg?
        # 这边所谓的score得分就是代表的有效连通分量的面积，太小的话就直接过滤掉代表无效的区域不要了
        instance_score_max = np.max(instance * output)    # score max
        if instance_score_max < cutoff_instance_max:
            continue
        instance_score_avg = np.sum(instance * output) / np.sum(instance)   # score avg
        if instance_score_avg < cutoff_instance_avg:
            continue
        filtered_instances.append(instance)
        scores.append(instance_score_avg) # 使用的是score_avg，用来最后算instance大小的
    instances = filtered_instances # 过滤掉太小的

    # dilation对二值图像进行膨胀操作，消除原来的空洞
    instances_tmp = []
    if post_dilation_iter > 0:
        for instance in filtered_instances:
            instance = ndimage.morphology.binary_dilation(instance, iterations=post_dilation_iter)
            instances_tmp.append(instance)
        instances = instances_tmp

    # sorted by size
    sorted_idx = [i[0] for i in sorted(enumerate(instances), key=lambda x: get_size_of_mask(x[1]))]
    instances = [instances[x] for x in sorted_idx]
    scores = [scores[x] for x in sorted_idx]

    # make sure there are no overlaps
    # todo: this dataset gt has overlap, so do not use this func
    # 移除重叠的部分
    instances, scores = remove_overlaps(instances, scores)

    # fill holes, 填补一些因为腐蚀而产生的洞
    if post_fill_holes:
        instances = [ndimage.morphology.binary_fill_holes(i) for i in instances]
    
    # instances = [np.expand_dims(i, axis=2) for i in instances]
    # scores = np.array(scores)
    # scores = np.expand_dims(scores, axis=1)
    if len(instances) > 0:
        lab_img = np.zeros(instances[0].shape, dtype=np.int32)
        for i, instance in enumerate(instances):
            lab_img = np.maximum(lab_img, instance * (i + 1))
        return lab_img
    else:
        return None

# 获取label图里面的instance
def getInstancePosition(gt_label,margin=12):
# image position => (y:y+h,x:x+w)
    max = np.max(gt_label)
    instanceList = []
    for idx in range(1,max):
        gt_copy = gt_label.copy()
        gt_copy[gt_copy != idx] = 0
        gt_copy = gt_copy.astype('uint8')
        x,y,w,h = cv2.boundingRect(gt_copy)
        # instanceList.append([y-margin,y+h+margin,x-margin,x+w+margin])
        instanceList.append([x,y,w,h])
    return np.array(instanceList)


# 未使用opencv的方法获取instance的坐标
def getInstanceArray(gt_label):
    # 所有的instance的大小
    predset  = np.unique(gt_label[gt_label>0])
    instanceList = []
    for ic in predset:
        instance = gt_label.copy()
        instance[ instance != ic ] = 0
        instance[ instance == ic ] = 1
        # 获取不为0的位置
        icx, icy = np.nonzero(instance)
        maxx = icx.max()
        maxy = icy.max()
        minx = icx.min()
        miny = icy.min()
        mx = np.round((maxx+minx)/2)
        my = np.round((maxy+miny)/2)
        halfsz = (np.max([(maxx-minx)/2, (maxy-miny)/2, 8])+12).astype(np.int16)
        # 图像实际的起始位置
        sx = np.round(mx - halfsz).astype(np.int16)
        sy = np.round(my - halfsz).astype(np.int16)
        ex = np.round(mx + halfsz + 1).astype(np.int16)
        ey = np.round(my + halfsz + 1).astype(np.int16)
        # 添加instance的数据的，获取的是方形的数据
        instanceList.append([sx,sy,ex,ey])
    return instanceList


# 根据二值图来获取instance的起止坐标
def getInstancePosi(binaryImg):
    # 获取不为0的位置
    icx, icy = np.nonzero(binaryImg)
    maxx = icx.max()
    maxy = icy.max()
    minx = icx.min()
    miny = icy.min()
    mx = np.round((maxx+minx)/2)
    my = np.round((maxy+miny)/2)
    halfsz = (np.max([(maxx - minx)/2, (maxy - miny)/2, 8])+12).astype(np.int16)
    # 图像实际的起始位置
    sx = np.round(mx - halfsz).astype(np.int16)
    sy = np.round(my - halfsz).astype(np.int16)
    ex = np.round(mx + halfsz + 1).astype(np.int16)
    ey = np.round(my + halfsz + 1).astype(np.int16)
    # 添加instance的数据的，获取的是方形的数据
    return sx,sy,ex,ey








import warnings
import numpy as np
from scipy.optimize import linear_sum_assignment

import cv2
import matplotlib.pyplot as plt

# The 'metrics.py' file is borrowed from the project of 'HoVer-Net', in which the name of this file is 'state_utils.py'
#####--------------------------Optimized for Speed
def get_fast_aji(true, pred):
    """
    AJI version distributed by MoNuSeg, has no permutation problem but suffered from 
    over-penalisation similar to DICE2

    Fast computation requires instance IDs are in contiguous orderding i.e [1, 2, 3, 4] 
    not [2, 3, 6, 10]. Please call `remap_label` before hand and `by_size` flag has no 
    effect on the result.
    """
    true = np.copy(true) # ? do we need this
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [None,]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)
    
    pred_masks = [None,]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)
    
    # prefill with value
    pairwise_inter = np.zeros([len(true_id_list) -1, 
                               len(pred_id_list) -1], dtype=np.float64)
    pairwise_union = np.zeros([len(true_id_list) -1, 
                               len(pred_id_list) -1], dtype=np.float64)

    # caching pairwise
    for true_id in true_id_list[1:]: # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0: # ignore
                continue # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            pairwise_inter[true_id-1, pred_id-1] = inter
            pairwise_union[true_id-1, pred_id-1] = total - inter
    #
    pairwise_iou = pairwise_inter / (pairwise_union + 1.0e-6)
    # pair of pred that give highest iou for each true, dont care 
    # about reusing pred instance multiple times
    paired_pred = np.argmax(pairwise_iou, axis=1)
    pairwise_iou = np.max(pairwise_iou, axis=1)
    # exlude those dont have intersection
    paired_true = np.nonzero(pairwise_iou > 0.0)[0]
    paired_pred = paired_pred[paired_true]
    # print(paired_true.shape, paired_pred.shape)
    overall_inter = (pairwise_inter[paired_true, paired_pred]).sum()
    overall_union = (pairwise_union[paired_true, paired_pred]).sum()
    #
    paired_true = (list(paired_true + 1)) # index to instance ID
    paired_pred = (list(paired_pred + 1))
    # add all unpaired GT and Prediction into the union
    unpaired_true = np.array([idx for idx in true_id_list[1:] if idx not in paired_true])
    unpaired_pred = np.array([idx for idx in pred_id_list[1:] if idx not in paired_pred])
    for true_id in unpaired_true:
        overall_union += true_masks[true_id].sum()
    for pred_id in unpaired_pred:
        overall_union += pred_masks[pred_id].sum()
    #
    aji_score = overall_inter / overall_union
    return aji_score
#####
def get_fast_aji_plus(true, pred):
    """
    AJI+, an AJI version with maximal unique pairing to obtain overall intersecion.
    Every prediction instance is paired with at most 1 GT instance (1 to 1) mapping, unlike AJI 
    where a prediction instance can be paired against many GT instances (1 to many).
    Remaining unpaired GT and Prediction instances will be added to the overall union.
    The 1 to 1 mapping prevents AJI's over-penalisation from happening.

    Fast computation requires instance IDs are in contiguous orderding i.e [1, 2, 3, 4] 
    not [2, 3, 6, 10]. Please call `remap_label` before hand and `by_size` flag has no 
    effect on the result.
    """
    true = np.copy(true) # ? do we need this
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [None,]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)
    
    pred_masks = [None,]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)
    
    # prefill with value
    pairwise_inter = np.zeros([len(true_id_list) -1, 
                               len(pred_id_list) -1], dtype=np.float64)
    pairwise_union = np.zeros([len(true_id_list) -1, 
                               len(pred_id_list) -1], dtype=np.float64)

    # caching pairwise
    for true_id in true_id_list[1:]: # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0: # ignore
                continue # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            pairwise_inter[true_id-1, pred_id-1] = inter
            pairwise_union[true_id-1, pred_id-1] = total - inter
    #
    pairwise_iou = pairwise_inter / (pairwise_union + 1.0e-6)
    #### Munkres pairing to find maximal unique pairing
    paired_true, paired_pred = linear_sum_assignment(-pairwise_iou)
    ### extract the paired cost and remove invalid pair 
    paired_iou = pairwise_iou[paired_true, paired_pred]
    # now select all those paired with iou != 0.0 i.e have intersection
    paired_true = paired_true[paired_iou > 0.0]
    paired_pred = paired_pred[paired_iou > 0.0]
    paired_inter = pairwise_inter[paired_true, paired_pred]
    paired_union = pairwise_union[paired_true, paired_pred]
    paired_true = (list(paired_true + 1)) # index to instance ID
    paired_pred = (list(paired_pred + 1))
    overall_inter = paired_inter.sum()
    overall_union = paired_union.sum()
    # add all unpaired GT and Prediction into the union
    unpaired_true = np.array([idx for idx in true_id_list[1:] if idx not in paired_true])
    unpaired_pred = np.array([idx for idx in pred_id_list[1:] if idx not in paired_pred])
    for true_id in unpaired_true:
        overall_union += true_masks[true_id].sum()
    for pred_id in unpaired_pred:
        overall_union += pred_masks[pred_id].sum()
    #
    aji_score = overall_inter / overall_union
    return aji_score
#####
def get_fast_pq(true, pred, match_iou=0.5):
    """
    `match_iou` is the IoU threshold level to determine the pairing between
    GT instances `p` and prediction instances `g`. `p` and `g` is a pair
    if IoU > `match_iou`. However, pair of `p` and `g` must be unique 
    (1 prediction instance to 1 GT instance mapping).

    If `match_iou` < 0.5, Munkres assignment (solving minimum weight matching
    in bipartite graphs) is caculated to find the maximal amount of unique pairing. 

    If `match_iou` >= 0.5, all IoU(p,g) > 0.5 pairing is proven to be unique and
    the number of pairs is also maximal.    
    
    Fast computation requires instance IDs are in contiguous orderding 
    i.e [1, 2, 3, 4] not [2, 3, 6, 10]. Please call `remap_label` beforehand 
    and `by_size` flag has no effect on the result.

    Returns:
        [dq, sq, pq]: measurement statistic

        [paired_true, paired_pred, unpaired_true, unpaired_pred]: 
                      pairing information to perform measurement
                    
    """
    assert match_iou >= 0.0, "Cant' be negative"
    
    true = np.copy(true)
    pred = np.copy(pred)
    true_id_list = list(np.unique(true))
    pred_id_list = list(np.unique(pred))

    true_masks = [None,]
    for t in true_id_list[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)
    
    pred_masks = [None,]
    for p in pred_id_list[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)

    # prefill with value
    pairwise_iou = np.zeros([len(true_id_list) -1, 
                             len(pred_id_list) -1], dtype=np.float64)

    # caching pairwise iou
    for true_id in true_id_list[1:]: # 0-th is background
        t_mask = true_masks[true_id]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        for pred_id in pred_true_overlap_id:
            if pred_id == 0: # ignore
                continue # overlaping background
            p_mask = pred_masks[pred_id]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            iou = inter / (total - inter)
            pairwise_iou[true_id-1, pred_id-1] = iou
    #
    if match_iou >= 0.5:
        paired_iou = pairwise_iou[pairwise_iou > match_iou]
        pairwise_iou[pairwise_iou <= match_iou] = 0.0
        paired_true, paired_pred = np.nonzero(pairwise_iou)
        paired_iou = pairwise_iou[paired_true, paired_pred]
        paired_true += 1 # index is instance id - 1
        paired_pred += 1 # hence return back to original
    else:  # * Exhaustive maximal unique pairing
        #### Munkres pairing with scipy library
        # the algorithm return (row indices, matched column indices)
        # if there is multiple same cost in a row, index of first occurence 
        # is return, thus the unique pairing is ensure
        # inverse pair to get high IoU as minimum   
        paired_true, paired_pred = linear_sum_assignment(-pairwise_iou)
        ### extract the paired cost and remove invalid pair 
        paired_iou = pairwise_iou[paired_true, paired_pred]

        # now select those above threshold level
        # paired with iou = 0.0 i.e no intersection => FP or FN
        paired_true = list(paired_true[paired_iou > match_iou] + 1)
        paired_pred = list(paired_pred[paired_iou > match_iou] + 1)
        paired_iou = paired_iou[paired_iou > match_iou]

    # get the actual FP and FN
    unpaired_true = [idx for idx in true_id_list[1:] if idx not in paired_true]
    unpaired_pred = [idx for idx in pred_id_list[1:] if idx not in paired_pred]
    # print(paired_iou.shape, paired_true.shape, len(unpaired_true), len(unpaired_pred))

    #
    tp = len(paired_true)
    fp = len(unpaired_pred)
    fn = len(unpaired_true)
    # get the F1-score i.e DQ
    dq = tp / (tp + 0.5 * fp + 0.5 * fn)
    # get the SQ, no paired has 0 iou so not impact
    sq = paired_iou.sum() / (tp + 1.0e-6)

    return [dq, sq, dq * sq], [paired_true, paired_pred, unpaired_true, unpaired_pred]

#####
def get_fast_dice_2(true, pred):
    """
        Ensemble dice
    """
    true = np.copy(true)
    pred = np.copy(pred)
    true_id = list(np.unique(true))
    pred_id = list(np.unique(pred))

    overall_total = 0
    overall_inter = 0

    true_masks = [np.zeros(true.shape)]
    for t in true_id[1:]:
        t_mask = np.array(true == t, np.uint8)
        true_masks.append(t_mask)
    
    pred_masks = [np.zeros(true.shape)]
    for p in pred_id[1:]:
        p_mask = np.array(pred == p, np.uint8)
        pred_masks.append(p_mask)
 
    for true_idx in range(1, len(true_id)):
        t_mask = true_masks[true_idx]
        pred_true_overlap = pred[t_mask > 0]
        pred_true_overlap_id = np.unique(pred_true_overlap)
        pred_true_overlap_id = list(pred_true_overlap_id)
        try: # blinly remove background
            pred_true_overlap_id.remove(0)
        except ValueError:
            pass  # just mean no background
        for pred_idx in pred_true_overlap_id:
            p_mask = pred_masks[pred_idx]
            total = (t_mask + p_mask).sum()
            inter = (t_mask * p_mask).sum()
            overall_total += total
            overall_inter += inter

    return 2 * overall_inter / overall_total
#####

def get_dice_1_multicls(true, pred, ths=0.0, eps=1e-7):
    """
        Traditional dice
    """
    assert(len(pred.shape)==3)
    ncls = pred.shape[0]
    dsc = 0.0
    for icls in range(1, ncls):
        ictrue = (true==icls).astype(np.float32)
        icpred = (pred[icls]>ths).astype(np.float32)
        inter = ictrue * icpred 
        denom = ictrue + icpred 
        dsc += (2.0 * np.sum(inter) + eps) / (np.sum(denom) + eps)
    dsc /= (ncls-1)
    return dsc
    

#####--------------------------As pseudocode
def get_dice_1(true, pred, ths=0.0):
    """
        Traditional dice
    """
    # cast to binary 1st
    true = np.copy(true)
    pred = np.copy(pred)
    true[true > ths] = 1
    pred[pred > ths] = 1
    inter = true * pred
    denom = true + pred
    return 2.0 * np.sum(inter) / np.sum(denom)

####
def get_dice_2(true, pred):
    true = np.copy(true)
    pred = np.copy(pred)
    true_id = list(np.unique(true))
    pred_id = list(np.unique(pred))
    # remove background aka id 0
    true_id.remove(0)
    pred_id.remove(0)

    total_markup = 0
    total_intersect = 0
    for t in true_id:
        t_mask = np.array(true == t, np.uint8)
        for p in pred_id:
            p_mask = np.array(pred == p, np.uint8)
            intersect = p_mask * t_mask          
            if intersect.sum() > 0:
                total_intersect += intersect.sum()
                total_markup += (t_mask.sum() + p_mask.sum())
    return 2 * total_intersect / total_markup
#####
def remap_label(pred, by_size=False):
    """
    Rename all instance id so that the id is contiguous i.e [0, 1, 2, 3] 
    not [0, 2, 4, 6]. The ordering of instances (which one comes first) 
    is preserved unless by_size=True, then the instances will be reordered
    so that bigger nucler has smaller ID

    Args:
        pred    : the 2d array contain instances where each instances is marked
                  by non-zero integer
        by_size : renaming with larger nuclei has smaller id (on-top)
    """
    pred_id = list(np.unique(pred))
    pred_id.remove(0)
    if len(pred_id) == 0:
        return pred # no label
    if by_size:
        pred_size = []
        for inst_id in pred_id:
            size = (pred == inst_id).sum()
            pred_size.append(size)
        # sort the id by size in descending order
        pair_list = zip(pred_id, pred_size)
        pair_list = sorted(pair_list, key=lambda x: x[1], reverse=True)
        pred_id, pred_size = zip(*pair_list)

    new_pred = np.zeros(pred.shape, np.int32)
    for idx, inst_id in enumerate(pred_id):
        new_pred[pred == inst_id] = idx + 1    
    return new_pred
#####
def pair_coordinates(setA, setB, radius):
    """
    Use the Munkres or Kuhn-Munkres algorithm to find the most optimal 
    unique pairing (largest possible match) when pairing points in set B 
    against points in set A, using distance as cost function

    Args:
        setA, setB: np.array (float32) of size Nx2 contains the of XY coordinate
                    of N different points 
        radius: valid area around a point in setA to consider 
                a given coordinate in setB a candidate for match
    Return:
        pairing: pairing is an array of indices
        where point at index pairing[0] in set A paired with point
        in set B at index pairing[1]
        unparedA, unpairedB: remaining poitn in set A and set B unpaired
    """

    # * Euclidean distance as the cost matrix
    setA_tile = np.expand_dims(setA, axis=1)
    setB_tile = np.expand_dims(setB, axis=0)
    setA_tile = np.repeat(setA_tile, setB.shape[0], axis=1)
    setB_tile = np.repeat(setB_tile, setA.shape[0], axis=0)
    pair_distance = (setA_tile - setB_tile) ** 2
    # set A is row, and set B is paired against set A
    pair_distance = np.sqrt(np.sum(pair_distance, axis=-1))

    # * Munkres pairing with scipy library
    # the algorithm return (row indices, matched column indices)
    # if there is multiple same cost in a row, index of first occurence 
    # is return, thus the unique pairing is ensured
    indicesA, paired_indicesB = linear_sum_assignment(pair_distance)

    # extract the paired cost and remove instances 
    # outside of designated radius
    pair_cost = pair_distance[indicesA, paired_indicesB]

    pairedA = indicesA[pair_cost <= radius]
    pairedB = paired_indicesB[pair_cost <= radius]

    unpairedA = [idx for idx in range(setA.shape[0]) if idx not in list(pairedA)]
    unpairedB = [idx for idx in range(setB.shape[0]) if idx not in list(pairedB)]

    pairing = np.array(list(zip(pairedA, pairedB)))    
    unpairedA = np.array(unpairedA, dtype=np.int64)
    unpairedB = np.array(unpairedB, dtype=np.int64)

    return pairing, unpairedA, unpairedB



import numpy as np

# 将数据集合分成k折
def kfold_list(n, K, seed=123):
    '''
    kfold_list
    Args:
        input: n,k,seed
        output: train_from_list, val_from_list
    '''
    rnd_state = np.random.RandomState(seed)
    dataset_rnd_sort = np.array(list(range(n)))
    rnd_state.shuffle(dataset_rnd_sort)
    idxlist = list(range(0, n, 1))
    val_idx = [list(range(i*(n//K), (i+1)*(n//K))) for i in range(0, K, 1)]
    train_idx = [list(range(0, i*int(n//K), 1))+list(range((i+1)*(n//K), n, 1)) for i in range(0, K, 1)]
    if val_idx[-1][-1] < n-1:
        val_idx[-1] = val_idx[-1]+list(range(val_idx[-1][-1]+1, n, 1))
        train_idx[-1] = list(range(0, (K-1)*(n//K)))
    val_from_list = []
    train_from_list = []
    for ival, itrain in zip(val_idx, train_idx):
        val_from_list.append(dataset_rnd_sort[ival])
        train_from_list.append(dataset_rnd_sort[itrain])
    return train_from_list, val_from_list


import setuptools


with open('README.md','r') as fh:
    long_description = fh.read()


setuptools.setup(
    name = 'JiweiCommonUtil',
    version = '0.0.3',
    author = '纪伟',
    author_email = '2413914266@qq.com',
    description = '自己常用的一些包，为了使用方便封装起来传到pypi',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JiweiMaster/jiwei_pypkg_tutorial",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import numpy as np
import scipy.io as scio
import cv2

# 染色归一化
def normalizeStaining(img, saveFile=None, Io=240, alpha=1, beta=0.15):
    ''' Normalize staining appearence of H&E stained images
    
    Example use:
        see test.py
        
    Input:
        I: RGB input image
        Io: (optional) transmitted light intensity
        
    Output:
        Inorm: normalized image
        H: hematoxylin image
        E: eosin image
    
    Reference: 
        A method for normalizing histology slides for quantitative analysis. M.
        Macenko et al., ISBI 2009
    '''
             
    HERef = np.array([[0.5626, 0.2159],
                      [0.7201, 0.8012],
                      [0.4062, 0.5581]])
        
    maxCRef = np.array([1.9705, 1.0308])
    
    # define height and width of image
    h, w, c = img.shape
    
    # reshape image
    img = img.reshape((-1,3))

    # calculate optical density
    OD = -np.log((img.astype(np.float)+1)/Io)
    
    # remove transparent pixels
    ODhat = OD[~np.any(OD<beta, axis=1)]
        
    # compute eigenvectors
    eigvals, eigvecs = np.linalg.eigh(np.cov(ODhat.T))
    
    #eigvecs *= -1
    
    #project on the plane spanned by the eigenvectors corresponding to the two 
    # largest eigenvalues    
    That = ODhat.dot(eigvecs[:,1:3])
    
    phi = np.arctan2(That[:,1],That[:,0])
    
    minPhi = np.percentile(phi, alpha)
    maxPhi = np.percentile(phi, 100-alpha)
    
    vMin = eigvecs[:,1:3].dot(np.array([(np.cos(minPhi), np.sin(minPhi))]).T)
    vMax = eigvecs[:,1:3].dot(np.array([(np.cos(maxPhi), np.sin(maxPhi))]).T)
    
    # a heuristic to make the vector corresponding to hematoxylin first and the 
    # one corresponding to eosin second
    if vMin[0] > vMax[0]:
        HE = np.array((vMin[:,0], vMax[:,0])).T
    else:
        HE = np.array((vMax[:,0], vMin[:,0])).T
    
    # rows correspond to channels (RGB), columns to OD values
    Y = np.reshape(OD, (-1, 3)).T
    
    # determine concentrations of the individual stains
    C = np.linalg.lstsq(HE,Y, rcond=None)[0]
    
    # normalize stain concentrations
    maxC = np.array([np.percentile(C[0,:], 99), np.percentile(C[1,:],99)])
    tmp = np.divide(maxC,maxCRef)
    C2 = np.divide(C,tmp[:, np.newaxis])
    
    # recreate the image using reference mixing matrix
    Inorm = np.multiply(Io, np.exp(-HERef.dot(C2)))
    Inorm[Inorm>255] = 254
    Inorm = np.reshape(Inorm.T, (h, w, 3)).astype(np.uint8)  
    
    # unmix hematoxylin and eosin
    H = np.multiply(Io, np.exp(np.expand_dims(-HERef[:,0], axis=1).dot(np.expand_dims(C2[0,:], axis=0))))
    H[H>255] = 254
    H = np.reshape(H.T, (h, w, 3)).astype(np.uint8)
    
    E = np.multiply(Io, np.exp(np.expand_dims(-HERef[:,1], axis=1).dot(np.expand_dims(C2[1,:], axis=0))))
    E[E>255] = 254
    E = np.reshape(E.T, (h, w, 3)).astype(np.uint8)
    
    if saveFile is not None:
        Image.fromarray(Inorm).save(saveFile+'.png')
        Image.fromarray(H).save(saveFile+'_H.png')
        Image.fromarray(E).save(saveFile+'_E.png')

    return Inorm, H, E


import matplotlib.pyplot as plt

'''
显示多张图片
'''


def showLineImg(imgList,figsize=(10,6),title=[]):
    length = len(imgList)
    fig,ax = plt.subplots(1,length,figsize = figsize)
    for i in range(length):
        ax[i].imshow(imgList[i])
        if len(title) == len(imgList):
            ax[i].set_title(title[i],fontsize=12,color='r')




